/*-------------------------------------------------------------------------
 *
 * dumpcache.c
 *	  disply some contents for the buffer cache
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"
#include "catalog/pg_type.h"
#include "access/heapam.h"
#include "miscadmin.h"
#include "nodes/parsenodes.h"
#include "storage/buf_internals.h"
#include "storage/bufmgr.h"
#include "storage/bufpage.h"
#include "utils/relcache.h"
#include "utils/hsearch.h"
#include "commands/prepare.h"
#include "commands/dumpcache.h"

/*
 *  DumpCacheResDesc
 *
 *  fake a tuple description for returning
 */
TupleDesc
DumpCacheResDesc(DumpCacheStmt *dumpstmt)
{

	TupleDesc	tupdesc;

	/* Just use a single TEXT column */
	tupdesc = CreateTemplateTupleDesc(1, false);
	TupleDescInitEntry(tupdesc, (AttrNumber) 1, "CACHE DUMP",
						TEXTOID, -1, 0, false);
 
	return tupdesc;

}


/*
 *	DumpCache 
 * 
 *  display the buffercache
 */
void
DumpCache(DumpCacheStmt *dumpstmt, DestReceiver *dest)
{
	int				i;
	BufferDesc		*bufHdr;
	RelFileNode		rnode;
	BlockNumber		blocknum;
	Relation		reln;
	char			*relname;
	char			relstate[10];
	HTAB			*relblockhash;
	HASHCTL			rctl;
	HASH_SEQ_STATUS rstatus;
	RelBlockEntry	*rentry;
	bool			found;
	TupOutputState	*tstate;
	char			*textbuf;
	

	elog(DEBUG1, "dumping summary of shared buffers (relation page counts):");


	/*
	 * initialize a hash for relations and block counts
	 */
	rctl.keysize = NAMEDATALEN + 1;
	rctl.entrysize = sizeof(RelBlockEntry);

	relblockhash = hash_create("Rel Block Hash", 
					100 , &rctl, HASH_ELEM);


	for (i = 0, bufHdr = BufferDescriptors; i < NBuffers; i++, bufHdr++)
	{

		/*
		 * get the relation node and block number from the buffer header
		 */
		rnode = bufHdr->tag.rnode;
		blocknum = bufHdr->tag.blockNum;


		/*
		 * assign a placeholder name for empty pages
		 * otherwise get the relation for this node
		 * handle the case where the rel metadata is not in the cache
		 * by loading the relation - handle when this fails (relation in 
		 * another database or dropped ?)
		 */
		if ( blocknum == InvalidBlockNumber )
		{
			relname = EMPTYRELNAME;
			strcpy(relstate, "norel");
		}
		else if (rnode.tblNode == InvalidOid || rnode.relNode == InvalidOid)
		{
			relname = INVALIDRELNAME;
			strcpy(relstate, "invalid");
		}
		else
		{
			reln = RelationNodeCacheGetRelation(rnode);


			if ( reln == NULL)
			{
				reln = RelationIdGetRelation(bufHdr->tag.rnode.relNode);
				if (reln == NULL)
				{
					relname = NOTTHISDBRELNAME;	/* another database ? */
					strcpy(relstate, "other");	
				}
				else
				{
					relname = RelationGetRelationName(reln);
					strcpy(relstate, "loaded");	/* valid relation */
				}
			}
			else
			{
		
				relname = RelationGetRelationName(reln);
				strcpy(relstate, "cached"); 
			}


			/*
			 * finished with the relation itself now
			 * but dont try to close if it's NULL
			 */
			if ( reln != NULL ) 
			{
				RelationClose(reln);
			}

		}


		/*
		 * show the buffer
		 */
		elog(DEBUG2, "  buffer %d - block %u of (%s) %s", 
			bufHdr->buf_id + 1,
			bufHdr->tag.blockNum,
			relstate,
			relname);


		/*
		 * search for this relation in the relation block hash
		 * and increment its count
		 * add an new entry if do not find
		 */

		elog(DEBUG2, "  processing entry for %s", relname);
		rentry = (RelBlockEntry *) hash_search(relblockhash, 
											(void *)relname, 
											HASH_FIND, NULL);
		if (rentry != NULL)
		{
			rentry->blocks++;
			elog(DEBUG2, "  updated");
		}
		else
		{
			rentry = (RelBlockEntry *) hash_search(relblockhash, 
												(void *)relname, 
												HASH_ENTER, &found);
			if (rentry == NULL) 
			{
				elog(ERROR, "DumpCache : out of memory for hash");
			}
			if (found) 
			{
				elog(ERROR, "DumpCache : hash found while entering");
			}
		
			strcpy(rentry->name, relname);
			rentry->blocks = 1;
		}


	}

	
	/*
	 * initialize the hash for sequential access
	 * and allocate a text buffer for psuedo tuple projection
	 */
	hash_seq_init(&rstatus, relblockhash);

	textbuf = (char *)palloc0(2 * NAMEDATALEN);

	tstate = begin_tup_output_tupdesc(dest, DumpCacheResDesc(dumpstmt));


	while ((rentry = (RelBlockEntry *)hash_seq_search(&rstatus)) != NULL)
	{
		sprintf(textbuf, "%-50s %12u", rentry->name, rentry->blocks);
		elog(DEBUG1, textbuf);
		do_text_output_multiline(tstate, textbuf);
	}


	end_tup_output(tstate);
	pfree(textbuf);

	hash_destroy(relblockhash);

	elog(DEBUG1, "end dump");


	return;

}
